{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE ViewPatterns      #-}

module Lib (splitIntoDocs, tokenize, pruneTrie, triesFrom, lineScore, lineComplete, 
  Trie, docNGrams, par') where

import           Control.DeepSeq             (NFData, rnf)
import           Control.Monad               (ap, join)
import           Control.Parallel.Strategies (parBuffer, rdeepseq, using)
import           Data.Char                   (isPunctuation, isSymbol)
import           Data.List                   (foldl', foldl1', intercalate,
                                              sortBy, unzip3, zip4)
import           Data.List.Split             (chunksOf, divvy, splitOn)
import qualified Data.Map.Strict             as M
import           Data.Maybe                  (mapMaybe)
import qualified Data.Text.Lazy              as T
import           System.Random               (randomRIO)

--- Types ---
data Trie = Trie Count (M.Map Token Trie)

instance Semigroup Trie where
  (Trie c1 m1) <> (Trie c2 m2) = Trie (c1 + c2) $ M.unionWith (<>) m1 m2

instance Monoid Trie where
  mempty = Trie 0 M.empty
  mconcat = foldl1' (<>)

instance NFData Trie where
  rnf (Trie c m) = rnf c `seq` rnf m

instance Show Trie where
  show t = show_ t 1
    where
      show_ (Trie c m) d
        | null m = show c
        | otherwise = "(" ++ show c ++ ", " ++ 
          dictShow (sortBy (\(_, Trie a _) (_, Trie b _) -> compare b a) (M.toList m)) d
          ++ ")"
      dictShow l d = "{" ++ intercalate ", " (map (\(k, v) -> "\n" ++
                replicate (d * 2) ' ' ++ show k ++ ": " ++ show_ v (d + 1))
              l) ++ "\n" ++ replicate ((d - 1) * 2) ' ' ++ "}"

-- Types for building the trie --
type Token = T.Text

type Line = [Token]

type NGram = [Token]

type Document = [Line]

type Corpus = [Document]

type UntokenizedLine = T.Text

type UntokenizedDocument = [UntokenizedLine]

type UntokenizedCorpus = [UntokenizedDocument]

-- Types for using the trie --
type Count = Int

type ApproxCount = Float

type Score = Float

type BeamState = (NGram, Score, Line)

type PotentialBeamState = (NGram, Score, Line, [(Token, Score)])

--- Constants ---
bosToken :: Token
bosToken = "<s>"

bufferSize :: Int
bufferSize = 100

chunkSize :: Int
chunkSize = 10

backoffWt :: Float
backoffWt = 0.4

penaltyWt :: Float
penaltyWt = 20

--- Utility functions ---
par' :: NFData a => [a] -> [a]
par' = (`using` parBuffer bufferSize rdeepseq)

--- Functions for processing text ---
splitIntoDocs :: UntokenizedLine -> [UntokenizedLine] -> UntokenizedCorpus
splitIntoDocs docSep = splitOn [docSep]

naiveNGrams :: Int -> Line -> [NGram]
naiveNGrams _ []       = []
naiveNGrams n l@(_:xs) = take n l : naiveNGrams n xs

nGrams :: Int -> Line -> [NGram]
nGrams n toks = naiveNGrams n (replicate (n - 1) bosToken ++ toks)

-- Adapted from Data.Text's split function to keep separators
splitKeepSeps :: (Char -> Bool) -> T.Text -> [T.Text]
splitKeepSeps _ t@(T.null -> True) = [t]
splitKeepSeps p t = loop t
  where
    loop s
      | T.null s' = [l]
      | otherwise = l : T.singleton (T.head s') : loop (T.tail s')
      where
        (l, s') = T.break p s

tokenize :: UntokenizedLine -> Line
tokenize = filter (/= T.empty) . splitSpecialChars . T.words . T.toLower
  where
    splitSpecialChars =
      concatMap $ splitKeepSeps $ oneOf [isPunctuation, isSymbol]
    oneOf ps = or . ap ps . return

--- Functions for building n-gram tries ---
docNGrams :: Int -> Document -> [NGram]
docNGrams = (. join) . nGrams

insertNGram :: Trie -> NGram -> Trie
insertNGram (Trie c m) [] = Trie (c + 1) m
insertNGram (Trie c m) (gram:grams) = Trie (c + 1) $ M.alter go gram m
  where
    go Nothing  = Just $ insertNGram mempty grams
    go (Just t) = Just $ insertNGram t grams

buildTrie :: [NGram] -> Trie
buildTrie = foldl' insertNGram mempty

allTriesFrom :: Int -> Corpus -> [Trie]
allTriesFrom n = par' . map (buildTrie . docNGrams n)

triesFrom :: Int -> Count -> Corpus -> [Trie]
triesFrom n = (par' .) . (. allTriesFrom n) . mapMaybe . pruneTrie

pruneTrie :: Count -> Trie -> Maybe Trie
pruneTrie gramThreshold (Trie c m)
  | c < gramThreshold = Nothing
  | otherwise = Just $ Trie c $ M.mapMaybe (pruneTrie gramThreshold) m

mergeTries :: [Trie] -> Trie
mergeTries [ts] = ts
mergeTries ts = mergeTries $ mergeTriesOnce ts
  where
    mergeTriesOnce = par' . map mconcat . chunksOf chunkSize

--- Functions for using n-gram tries to compute probability of text ---
-- (Count of w_i...w_i+n-1, Count of w_i...w_i+n-2)
counts :: NGram -> Count -> Maybe Trie -> (Count, Count)
counts _ prev_c Nothing                 = (0, prev_c)
counts [] prev_c (Just (Trie c _))      = (c, prev_c)
counts (gram:grams) _ (Just (Trie c m)) = counts grams c $ M.lookup gram m

countsStupidBackoff :: NGram -> Maybe Trie -> (ApproxCount, ApproxCount)
countsStupidBackoff [] _ = (backoffWt ^ (100 :: Int), 1) -- Word not seen in training
countsStupidBackoff ngram trie
  | c == 0 =
    let (c_bo, prev_c_bo) = countsStupidBackoff (tail ngram) trie -- Back off stupidly
     in (backoffWt * c_bo, prev_c_bo)
  | otherwise = (fromIntegral c, fromIntegral prev_c)
  where
    (c, prev_c) = counts ngram 0 trie

nGramScore :: [Trie] -> NGram -> Score
nGramScore tries ngram =
  let (c, prev_c) =
        sumTuples (par' $ map (countsStupidBackoff ngram . pure) tries)
   in log c - log prev_c
  where
    sumTuples = foldl1' (\(a1, b1) (a2, b2) -> (a1 + a2, b1 + b2))

lineScore :: [Trie] -> Int -> UntokenizedLine -> Score
lineScore tries n =
  sum . par' . map (nGramScore tries) . divvy n 1 . 
  (replicate (n - 1) bosToken ++) . tokenize

--- Functions for using n-gram tries to generate text ---
trieFind :: NGram -> Maybe Trie -> Maybe Trie
trieFind [] t                           = t
trieFind _ Nothing                      = Nothing
trieFind (gram:grams) (Just (Trie _ m)) = trieFind grams $ M.lookup gram m

trieFindMerge :: NGram -> [Trie] -> Trie
trieFindMerge n = mergeTries . par' . mapMaybe (trieFind n . Just)

topBToks :: Int -> Trie -> [(Token, Score)]
topBToks b (Trie c m) =
  map (\(tok, Trie c1 _) -> (tok, log (fromIntegral c1) - log (fromIntegral c))) $
  take b $ sortBy (\(_, Trie c1 _) (_, Trie c2 _) -> compare c2 c1) $ M.toList m

-- Penalize with Hamming distance, number of shared (ordered) tokens between two lines
-- sim("Functional programming is cool and advanced", "Advanced Programming is boring") = 2
diversify :: [BeamState] -> [(BeamState, Score)]
diversify = f []
  where
    f _ [] = []
    f hs (x@(_, _, h):xs) =
      (x, penaltyWt * fromIntegral (sim h hs)) : f (h : hs) xs
    sim _ [] = 0
    sim h hs =
      length $ filter id $ foldl1' (zipWith (||)) $ map (zipWith (==) h) hs

-- Each current state has b candidate expansions
-- We define state as a tuple of score, history, current n gram, and list of expansions
-- Process each expansion into its own state by merging its information with parent state
scoreCandidates :: [PotentialBeamState] -> [(BeamState, Score)]
scoreCandidates = diversify . concatMap f
  where
    f (n, s, h, cs) = map (\(tok, s') -> (tail n ++ [tok], s + s', tok : h)) cs

beamSearch :: Int -> [Trie] -> [NGram] -> [Score] -> [Line] -> UntokenizedLine
 -> IO String
beamSearch b tries ngrams scores hists input = do
  randExitFlag <- randomRIO (0, 100 :: Int)
  let histStr = unlines $ map (T.unpack . T.unwords . (input :) . reverse) hists
      beamTries = par' $ map ((`trieFindMerge` tries) . tail) ngrams
      candidates = par' $ map (topBToks b) beamTries
      scoredCands = scoreCandidates $ zip4 ngrams scores hists candidates
      sortedCands = sortBy finalScoresAsc scoredCands
      (ngrams', scores', hists') = unzip3 $ map fst $ take b sortedCands
  if randExitFlag == 0
    then return histStr
  else beamSearch b tries ngrams' scores' hists' input
  where finalScoresAsc ((_, s1, _), p1) ((_, s2, _), p2) = compare (s2 - p2) (s1 - p1)

wordAtIndex :: Int -> [(Token, Trie)] -> Token
wordAtIndex _ [] = "" -- Couldn't find any child word (i.e. method was called on leaf)
wordAtIndex i ((tok, Trie c _):rest)
  | i <= 0 = tok
  | otherwise = wordAtIndex (i - c) rest

randomSearchTok :: NGram -> [Trie] -> IO Token
randomSearchTok ngram tries = do
  let (Trie c m) = trieFindMerge ngram tries
  idx <- randomRIO (0, c - 1)
  let tok = wordAtIndex idx $ M.toList m
  if T.null tok -- Couldn't find next word, back off stupidly!
    then randomSearchTok (tail ngram) tries
  else return tok

randomSearch :: [Trie] -> NGram -> IO String
randomSearch _ [] = error "Random search failed" -- Not reachable
randomSearch tries (_:grams) = do
  randExitFlag <- randomRIO (0, 2 :: Int)
  if randExitFlag == 0
    then return "..."
  else do
    tok <- randomSearchTok grams tries
    rest <- randomSearch tries (grams ++ [tok])
    return $ T.unpack tok ++ " " ++ rest

lineComplete :: String -> [Trie] -> Int -> UntokenizedLine -> IO String
lineComplete mode tries n line = do
  let lastNGram =
        (last . divvy n 1 . (replicate (n - 1) bosToken ++) . tokenize) line
      launchBeamSearch b = beamSearch b tries [lastNGram] [0] [[]] line
  case mode of
    "beamSearch" -> launchBeamSearch 5
    "greedy" -> launchBeamSearch 1
    "random" -> do
      putStr $ T.unpack line ++ " "
      randomSearch tries lastNGram
    _ ->
      error $ "Mode " ++ mode ++ 
      " is unsupported. Please choose from [beamSearch, greedy, random]."
